/*
 *  Utility functions to find cubic roots,
 *  coefficients are passed like this:
 *
 *      c[0] + c[1]*x + c[2]*x^2 + c[3]*x^3  = 0
 *
 *  The functions return the number of non-complex roots and
 *  put the values into the s array.
 *
 *  Author:         Jochen Schwarze (schwarze@isa.de)
 *
 *  Jan 26, 1990    Version for Graphics Gems
 *  Oct 11, 1990    Fixed sign problem for negative q's in SolveQuartic
 *  	    	    (reported by Mark Podlipec),
 *  	    	    Old-style function definitions,
 *  	    	    IsZero() as a macro
 *  Nov 23, 1990    Some systems do not declare acos() and cbrt() in
 *                  <math.h>, though the functions exist in the library.
 *                  If large coefficients are used, EQN_EPS should be
 *                  reduced considerably (e.g. to 1E-30), results will be
 *                  correct but multiple roots might be reported more
 *                  than once.
 *
 *  Aaron Knoll note: this is code really requires double precision.
 *  This enormously affects speed. In general, this should ONLY
 *  be used as reference code.
 */

#ifndef __MANTA_CORE_CUBICSOLVER_H__
#define __MANTA_CORE_CUBICSOLVER_H__

#include <math.h>
#include <stdio.h>
#include <Core/Math/Trig.h>
#include <Core/Math/Expon.h>

namespace Manta
{

#define     EQN_EPS     1e-30		
#define	    IsZero(x)	((x) > -EQN_EPS && (x) < EQN_EPS)

#ifdef NOCBRT
#define     cbrt(x)     ((x) > 0.0 ? pow((x), 1.0/3.0) : \
                        ((x) < 0.0 ? -pow(-(x), 1.0/3.0) : 0.0))
#endif

    inline int SolveLinear(double *c, double *s) {
        if (c[ 1 ] < 0.0000001 && c[ 1 ] > -0.0000001)
             return 0;
        s[0] = -c[0]/c[1]; 
        return 1;
    }

    inline int SolveQuadratic(double *c, double *s) {
       double  A, B, C;
       if (c[ 2 ] < 0.0000001 && c[ 2 ] > -0.0000001)
          return SolveLinear(c, s);

       /* Ax^2 + Bx + C = 0 */
       A = c[2];
       B = c[1];
       C = c[0];
         
       double disc = B*B - 4*A*C;
       if (disc < 0) return 0;

       disc = Sqrt(disc);
       s[0] = (-B - disc) / (2*A);
       s[1] = (-B + disc) / (2*A);
       return 2;
    }

    inline int SolveCubic(double c[4], double s[3])
    {
        int     i, num;
        double  sub;
        double  A, B, C;
        double  sq_A, p, q;
        double  cb_p, D;

        if (c[ 3 ] < 0.0000001 && c[ 3 ] > -0.0000001)
            return SolveQuadratic(c, s);

        /* normal form: x^3 + Ax^2 + Bx + C = 0 */

        A = c[ 2 ] / c[ 3 ];
        B = c[ 1 ] / c[ 3 ];
        C = c[ 0 ] / c[ 3 ];

        /*  substitute x = y - A/3 to eliminate quadric term:
        x^3 +px + q = 0 */

        sq_A = A * A;
        p = 1.0/3 * (- 1.0/3 * sq_A + B);
        q = 1.0/2 * (2.0/27 * A * sq_A - 1.0/3 * A * B + C);

        /* use Cardano's formula */

        cb_p = p * p * p;
        D = q * q + cb_p;

        if (IsZero(D))
        {
            if (IsZero(q)) /* one triple solution */
            {
                s[ 0 ] = 0;
                num = 1;
            }
            else /* one single and one double solution */
            {
                double u = Cbrt(-q);
                s[ 0 ] = 2 * u;
                s[ 1 ] = - u;
                num = 2;
            }
        }
        else if (D < 0) /* Casus irreducibilis: three real solutions */
        {
            double phi = 1.0/3 * Acos(-q / Sqrt(-cb_p));
            double t = 2 * Sqrt(-p);
        
            s[ 0 ] =   t * Cos(phi);
            s[ 1 ] = - t * Cos(phi + M_PI / 3);
            s[ 2 ] = - t * Cos(phi - M_PI / 3);
            num = 3;
        }
        else /* one real solution */
        {
            double sqrt_D = Sqrt(D);
            double u = Cbrt(sqrt_D - q);
            double v = - Cbrt(sqrt_D + q);
        
            s[ 0 ] = u + v;
            num = 1;
        }

        /* resubstitute */

        sub = 1.0/3 * A;

        for (i = 0; i < num; ++i)
        s[ i ] -= sub;

        return num;
    }
};

#endif
